<?xml version="1.0" encoding="UTF-8"?>
<xsl:stylesheet version="1.0"    
    xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
    xmlns:msxsl="urn:schemas-microsoft-com:xslt">

<!-- (c) 2021, Trimble Inc. All rights reserved.                                               -->
<!-- Permission is hereby granted to use, copy, modify, or distribute this style sheet for any -->
<!-- purpose and without fee, provided that the above copyright notice appears in all copies   -->
<!-- and that both the copyright notice and the limited warranty and restricted rights notice  -->
<!-- below appear in all supporting documentation.                                             -->

<!-- TRIMBLE INC. PROVIDES THIS STYLE SHEET "AS IS" AND WITH ALL FAULTS.                       -->
<!-- TRIMBLE INC. SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTY OF MERCHANTABILITY               -->
<!-- OR FITNESS FOR A PARTICULAR USE. TRIMBLE INC. DOES NOT WARRANT THAT THE                   -->
<!-- OPERATION OF THIS STYLE SHEET WILL BE UNINTERRUPTED OR ERROR FREE.                        -->

<xsl:output method="xml" omit-xml-declaration="no" encoding="utf-8"/>

<!-- Set the numeric display details i.e. decimal point, thousands separator etc -->
<xsl:variable name="DecPt" select="'.'"/>    <!-- Change as appropriate for US/European -->
<xsl:variable name="GroupSep" select="','"/> <!-- Change as appropriate for US/European -->
<!-- Also change decimal-separator & grouping-separator in decimal-format below 
     as appropriate for US/European output -->
<xsl:decimal-format name="Standard" 
                    decimal-separator="."
                    grouping-separator=","
                    infinity="Infinity"
                    minus-sign="-"
                    NaN=""
                    percent="%"
                    per-mille="&#2030;"
                    zero-digit="0" 
                    digit="#" 
                    pattern-separator=";" />

<xsl:variable name="DecPl0" select="'#0'"/>
<xsl:variable name="DecPl1" select="concat('#0', $DecPt, '0')"/>
<xsl:variable name="DecPl2" select="concat('#0', $DecPt, '00')"/>
<xsl:variable name="DecPl3" select="concat('#0', $DecPt, '000')"/>
<xsl:variable name="DecPl4" select="concat('#0', $DecPt, '0000')"/>
<xsl:variable name="DecPl5" select="concat('#0', $DecPt, '00000')"/>
<xsl:variable name="DecPl6" select="concat('#0', $DecPt, '000000')"/>
<xsl:variable name="DecPl8" select="concat('#0', $DecPt, '00000000')"/>
<xsl:variable name="DecPl10" select="concat('#0', $DecPt, '0000000000')"/>

<xsl:variable name="fileExt" select="'gml'"/>

<!-- User variable definitions - Appropriate fields are displayed on the       -->
<!-- Survey Controller screen to allow the user to enter specific values       -->
<!-- which can then be used within the style sheet definition to control the   -->
<!-- output data.                                                              -->
<!--                                                                           -->
<!-- All user variables must be identified by a variable element definition    -->
<!-- named starting with 'userField' (case sensitive) followed by one or more  -->
<!-- characters uniquely identifying the user variable definition.             -->
<!--                                                                           -->
<!-- The text within the 'select' field for the user variable description      -->
<!-- references the actual user variable and uses the '|' character to         -->
<!-- separate the definition details into separate fields as follows:          -->
<!-- For all user variables the first field must be the name of the user       -->
<!-- variable itself (this is case sensitive) and the second field is the      -->
<!-- prompt that will appear on the Survey Controller screen.                  -->
<!-- The third field defines the variable type - there are four possible       -->
<!-- variable types: Double, Integer, String and StringMenu.  These variable   -->
<!-- type references are not case sensitive.                                   -->
<!-- The fields that follow the variable type change according to the type of  -->
<!-- variable as follow:                                                       -->
<!-- Double and Integer: Fourth field = optional minimum value                 -->
<!--                     Fifth field = optional maximum value                  -->
<!--   These minimum and maximum values are used by the Survey Controller for  -->
<!--   entry validation.                                                       -->
<!-- String: No further fields are needed or used.                             -->
<!-- StringMenu: Fourth field = number of menu items                           -->
<!--             Remaining fields are the actual menu items - the number of    -->
<!--             items provided must equal the specified number of menu items. -->
<!--                                                                           -->
<!-- The style sheet must also define the variable itself, named according to  -->
<!-- the definition.  The value within the 'select' field will be displayed in -->
<!-- the Survey Controller as the default value for the item.                  -->

<!-- **************************************************************** -->
<!-- Set global variables from the Environment section of JobXML file -->
<!-- **************************************************************** -->
<xsl:variable name="DistUnit"   select="/JOBFile/Environment/DisplaySettings/DistanceUnits" />
<xsl:variable name="AngleUnit"  select="/JOBFile/Environment/DisplaySettings/AngleUnits" />
<xsl:variable name="CoordOrder" select="/JOBFile/Environment/DisplaySettings/CoordinateOrder" />
<xsl:variable name="TempUnit"   select="/JOBFile/Environment/DisplaySettings/TemperatureUnits" />
<xsl:variable name="PressUnit"  select="/JOBFile/Environment/DisplaySettings/PressureUnits" />

<!-- Setup conversion factor for coordinate and distance values -->
<!-- Dist/coord values in JobXML file are always in metres -->
<xsl:variable name="DistConvFactor">
  <xsl:choose>
    <xsl:when test="$DistUnit='Metres'">1.0</xsl:when>
    <xsl:when test="$DistUnit='InternationalFeet'">3.280839895</xsl:when>
    <xsl:when test="$DistUnit='USSurveyFeet'">3.2808333333357</xsl:when>
    <xsl:otherwise>1.0</xsl:otherwise>
  </xsl:choose>
</xsl:variable>

<!-- Setup conversion factor for angular values -->
<!-- Angular values in JobXML file are always in decimal degrees -->
<xsl:variable name="AngleConvFactor">
  <xsl:choose>
    <xsl:when test="$AngleUnit='DMSDegrees'">1.0</xsl:when>
    <xsl:when test="$AngleUnit='Gons'">1.111111111111</xsl:when>
    <xsl:when test="$AngleUnit='Mils'">17.77777777777</xsl:when>
    <xsl:otherwise>1.0</xsl:otherwise>
  </xsl:choose>
</xsl:variable>

<!-- Setup boolean variable for coordinate order -->
<xsl:variable name="NECoords">
  <xsl:choose>
    <xsl:when test="$CoordOrder='North-East-Elevation'">true</xsl:when>
    <xsl:when test="$CoordOrder='X-Y-Z'">true</xsl:when>
    <xsl:otherwise>false</xsl:otherwise>
  </xsl:choose>
</xsl:variable>

<!-- Setup conversion factor for pressure values -->
<!-- Pressure values in JobXML file are always in millibars (hPa) -->
<xsl:variable name="PressConvFactor">
  <xsl:choose>
    <xsl:when test="$PressUnit='MilliBar'">1.0</xsl:when>
    <xsl:when test="$PressUnit='InchHg'">0.029529921</xsl:when>
    <xsl:when test="$PressUnit='mmHg'">0.75006</xsl:when>
    <xsl:otherwise>1.0</xsl:otherwise>
  </xsl:choose>
</xsl:variable>

<xsl:variable name="emulator">false</xsl:variable>
<xsl:variable name="emulatorVersion">2021.00</xsl:variable>
<xsl:variable name="fxlFileFolder">
  <xsl:choose>
    <xsl:when test="$emulator = 'true'">
      <xsl:value-of select="concat('C:\ProgramData\Trimble\Trimble Access Emulator ', $emulatorVersion, '\Trimble Data\System Files\')"/>
    </xsl:when>
    <xsl:otherwise>C:\ProgramData\Trimble\Trimble Data\System Files\</xsl:otherwise>
  </xsl:choose>  
</xsl:variable>  
<xsl:variable name="fxlFileName" select="concat($fxlFileFolder, /JOBFile/Environment/AssociatedFiles/FeatureCodeLibrary)"/>

<!-- Use the document() function to load the referenced FXL file as  -->
<!-- a node-set variable that will provide the feature code details. -->
<xsl:variable name="initFxlDefinition" select="document(string($fxlFileName))"/>

<!-- Build into a more manageable node set -->
<xsl:variable name="fxlDefinition">
  <xsl:element name="FeatureCodingDefinitions">
    <xsl:copy-of select="msxsl:node-set($initFxlDefinition)/child::*[1]/@*" />
    <xsl:call-template name="CopyOf">
      <xsl:with-param name="select" select="msxsl:node-set($initFxlDefinition)/child::*[1]"/>
    </xsl:call-template>
  </xsl:element>
</xsl:variable>

<xsl:template name="CopyOf">   <!-- Copies all the elements from the FXL file without any namespace references -->
  <xsl:param name="select" />

  <xsl:for-each select="$select">
    <xsl:for-each select="node()|text()">
      <xsl:choose>
        <xsl:when test="self::*">   <!-- Node -->
          <xsl:element name="{local-name()}">
            <xsl:copy-of select="@*" />
            <xsl:call-template name="CopyOf">
              <xsl:with-param name="select" select="." />
            </xsl:call-template>
          </xsl:element>
        </xsl:when>

        <xsl:otherwise>  <!-- Text -->
          <xsl:if test="not(self::comment())">  <!-- Skip any comment elements -->
            <xsl:value-of select="." />
          </xsl:if>
        </xsl:otherwise>
      </xsl:choose>
    </xsl:for-each>
  </xsl:for-each>
</xsl:template>

<xsl:variable name="pointCodes">
  <xsl:for-each select="msxsl:node-set($fxlDefinition)/FeatureCodingDefinitions/FeatureDefinitions/PointFeatureDefinition">
    <xsl:element name="code" namespace="">
      <xsl:element name="name" namespace="">
        <xsl:value-of select="@Code"/>
      </xsl:element>
      <xsl:element name="attributes" namespace="">
        <xsl:for-each select="Attributes/*">
          <xsl:element name="attribute" namespace="">
            <xsl:attribute name="name">
              <xsl:value-of select="@Name"/>
            </xsl:attribute>
            <xsl:attribute name="description">
              <xsl:value-of select="@Description"/>
            </xsl:attribute>
            <xsl:attribute name="descWithoutName">  <!-- The 'path' leading to the attribute name -->
              <xsl:if test="contains(@Description, '.')">
                <xsl:value-of select="substring(@Description, 1, string-length(@Description) - string-length(@Name) - 1)"/>
              </xsl:if>
            </xsl:attribute>
            <xsl:attribute name="type">
              <xsl:value-of select="name(.)"/>
            </xsl:attribute>
            <xsl:variable name="inPath">
              <xsl:call-template name="ExtractPathElements">
                <xsl:with-param name="pathString" select="@Description"/>
              </xsl:call-template>
            </xsl:variable>
            <xsl:element name="inPath" namespace="">
              <xsl:copy-of select="msxsl:node-set($inPath)/*"/>
            </xsl:element>
            <xsl:element name="outPath" namespace="">
              <xsl:call-template name="ReversedNodeSet">
                <xsl:with-param name="originalNodeSet" select="msxsl:node-set($inPath)"/>
                <xsl:with-param name="count" select="count(msxsl:node-set($inPath)/*)"/>
                <xsl:with-param name="item" select="count(msxsl:node-set($inPath)/*)"/>
              </xsl:call-template>
            </xsl:element>
          </xsl:element>
        </xsl:for-each>
      </xsl:element>
    </xsl:element>
  </xsl:for-each>
</xsl:variable>

<xsl:variable name="lineCodes">
  <xsl:for-each select="msxsl:node-set($fxlDefinition)/FeatureCodingDefinitions/FeatureDefinitions/LineFeatureDefinition">
    <xsl:element name="code" namespace="">
      <xsl:element name="name" namespace="">
        <xsl:value-of select="@Code"/>
      </xsl:element>
      <xsl:element name="attributes" namespace="">
        <xsl:for-each select="Attributes/*">
          <xsl:element name="attribute" namespace="">
            <xsl:attribute name="name">
              <xsl:value-of select="@Name"/>
            </xsl:attribute>
            <xsl:attribute name="description">
              <xsl:value-of select="@Description"/>
            </xsl:attribute>
            <xsl:attribute name="descWithoutName">  <!-- The 'path' leading to the attribute name -->
              <xsl:if test="contains(@Description, '.')">
                <xsl:value-of select="substring(@Description, 1, string-length(@Description) - string-length(@Name) - 1)"/>
              </xsl:if>
            </xsl:attribute>
            <xsl:attribute name="type">
              <xsl:value-of select="name(.)"/>
            </xsl:attribute>
            <xsl:variable name="inPath">
              <xsl:call-template name="ExtractPathElements">
                <xsl:with-param name="pathString" select="@Description"/>
              </xsl:call-template>
            </xsl:variable>
            <xsl:element name="inPath" namespace="">
              <xsl:copy-of select="msxsl:node-set($inPath)/*"/>
            </xsl:element>
            <xsl:element name="outPath" namespace="">
              <xsl:call-template name="ReversedNodeSet">
                <xsl:with-param name="originalNodeSet" select="msxsl:node-set($inPath)"/>
                <xsl:with-param name="count" select="count(msxsl:node-set($inPath)/*)"/>
                <xsl:with-param name="item" select="count(msxsl:node-set($inPath)/*)"/>
              </xsl:call-template>
            </xsl:element>
          </xsl:element>
        </xsl:for-each>
      </xsl:element>
    </xsl:element>
  </xsl:for-each>
</xsl:variable>

<xsl:variable name="controlCodes">
  <xsl:for-each select="msxsl:node-set($fxlDefinition)/FeatureCodingDefinitions/ControlCodeDefinitions/ControlCodeDefinition">
    <xsl:element name="code" namespace="">
      <xsl:attribute name="type">
        <xsl:value-of select="@Type"/>
      </xsl:attribute>
      <xsl:value-of select="@Code"/>
    </xsl:element>
  </xsl:for-each>
</xsl:variable>

<xsl:variable name="coordRange">
  <xsl:variable name="pointsSortN">
    <xsl:for-each select="/JOBFile/Reductions/Point[Grid/North != '']">
      <xsl:sort data-type="number" order="descending" select="Grid/North"/>
      <xsl:copy>
        <xsl:copy-of select="Grid/North" />
      </xsl:copy>
    </xsl:for-each>
  </xsl:variable>

  <xsl:element name="maxNorth" namespace="">
    <xsl:value-of select="msxsl:node-set($pointsSortN)/Point[1]/North + 5"/>
  </xsl:element>

  <xsl:element name="minNorth" namespace="">
    <xsl:value-of select="msxsl:node-set($pointsSortN)/Point[last()]/North - 5"/>
  </xsl:element>

  <xsl:variable name="pointsSortE">
    <xsl:for-each select="/JOBFile/Reductions/Point[Grid/East != '']">
      <xsl:sort data-type="number" order="descending" select="Grid/East"/>
      <xsl:copy>
        <xsl:copy-of select="Grid/East" />
      </xsl:copy>
    </xsl:for-each>
  </xsl:variable>

  <xsl:element name="maxEast" namespace="">
    <xsl:value-of select="msxsl:node-set($pointsSortE)/Point[1]/East + 5"/>
  </xsl:element>

  <xsl:element name="minEast" namespace="">
    <xsl:value-of select="msxsl:node-set($pointsSortE)/Point[last()]/East - 5"/>
  </xsl:element>

  <xsl:variable name="pointsSortElev">
    <xsl:for-each select="/JOBFile/Reductions/Point[Grid/Elevation != '']">
      <xsl:sort data-type="number" order="descending" select="Grid/Elevation"/>
      <xsl:copy>
        <xsl:copy-of select="Grid/Elevation" />
      </xsl:copy>
    </xsl:for-each>
  </xsl:variable>

  <xsl:element name="maxElev" namespace="">
    <xsl:value-of select="msxsl:node-set($pointsSortElev)/Point[1]/Elevation + 1"/>
  </xsl:element>

  <xsl:element name="minElev" namespace="">
    <xsl:choose>
      <xsl:when test="msxsl:node-set($pointsSortElev)/Point[last()]/Elevation &lt; -100">0</xsl:when>
      <xsl:otherwise>
        <xsl:value-of select="msxsl:node-set($pointsSortElev)/Point[last()]/Elevation - 1"/>
      </xsl:otherwise>
    </xsl:choose>
  </xsl:element>
  
  <xsl:element name="hasNullElevations" namespace="">
    <xsl:choose>
      <xsl:when test="count(/JOBFile/Reductions/Point[Grid/Elevation = '']) != 0">true</xsl:when>
      <xsl:otherwise>false</xsl:otherwise>
    </xsl:choose>
  </xsl:element>
</xsl:variable>

<xsl:variable name="environmentElement">
  <xsl:element name="Environment" namespace="">
    <xsl:for-each select="/JOBFile/Environment">
      <xsl:copy-of select="* | @*"/>
    </xsl:for-each>
  </xsl:element>
</xsl:variable>

<!-- The following is a list of the attributes that need to be output -->  
<!-- as sub-elements under prefix elements. The prefix elements are   --> 
<!-- included under an element with the name of the attribute.        -->  
<xsl:variable name="attribPrefixes">
  <xsl:element name="målemetode">
    <xsl:element name="prefix">kvalitet</xsl:element>
    <xsl:element name="prefix">Posisjonskvalitet</xsl:element>
  </xsl:element>
  <xsl:element name="nøyaktighet">
    <xsl:element name="prefix">kvalitet</xsl:element>
    <xsl:element name="prefix">Posisjonskvalitet</xsl:element>
  </xsl:element>
  <xsl:element name="målemetodeHøyde">
    <xsl:element name="prefix">kvalitet</xsl:element>
    <xsl:element name="prefix">Posisjonskvalitet</xsl:element>
  </xsl:element>
  <xsl:element name="nøyaktighetHøyde">
    <xsl:element name="prefix">kvalitet</xsl:element>
    <xsl:element name="prefix">Posisjonskvalitet</xsl:element>
  </xsl:element>
  <xsl:element name="produkt">
    <xsl:element name="prefix">registreringsversjon</xsl:element>
    <xsl:element name="prefix">Registreringsversjon</xsl:element>
  </xsl:element>
  <xsl:element name="versjon">
    <xsl:element name="prefix">registreringsversjon</xsl:element>
    <xsl:element name="prefix">Registreringsversjon</xsl:element>
  </xsl:element>
  <xsl:element name="lokalId">
    <xsl:element name="prefix">identifikasjon</xsl:element>
    <xsl:element name="prefix">Identifikasjon</xsl:element>
  </xsl:element>
  <xsl:element name="navnerom">
    <xsl:element name="prefix">identifikasjon</xsl:element>
    <xsl:element name="prefix">Identifikasjon</xsl:element>
  </xsl:element>
  <xsl:element name="eiertype">
    <xsl:element name="prefix">eier</xsl:element>
    <xsl:element name="prefix">Eier</xsl:element>
  </xsl:element>
  <xsl:element name="eier">
    <xsl:element name="prefix">eier</xsl:element>
    <xsl:element name="prefix">Eier</xsl:element>
  </xsl:element>
  <xsl:element name="eierAndel">
    <xsl:element name="prefix">eier</xsl:element>
    <xsl:element name="prefix">Eier</xsl:element>
  </xsl:element>
  <xsl:element name="maksAvvikVertikalt">
    <xsl:element name="prefix">maks3dAvvik</xsl:element>
    <xsl:element name="prefix">Maks3dAvvik</xsl:element>
  </xsl:element>
  <xsl:element name="maksAvvikHorisontalt">
    <xsl:element name="prefix">maks3dAvvik</xsl:element>
    <xsl:element name="prefix">Maks3dAvvik</xsl:element>
  </xsl:element>
</xsl:variable>


<!-- **************************************************************** -->
<!-- ************************** Main Loop *************************** -->
<!-- **************************************************************** -->
<xsl:template match="/" >

  <!-- Build up the appropriate schema reference from the schema name read from the FXL file -->
  <xsl:variable name="schema" select="substring-after(msxsl:node-set($fxlDefinition)/FeatureCodingDefinitions/@Description, '=')"/>

  <xsl:variable name="posnOfLastSlashCharInSchema">
    <xsl:call-template name="PosnOfLastOfSpecifiedChars">
      <xsl:with-param name="stringToSearch" select="$schema"/>
      <xsl:with-param name="charToLocate">/</xsl:with-param>
    </xsl:call-template>
  </xsl:variable>

  <xsl:variable name="schemaPath">  <!-- Has the 'path' to the schema followed by the full schema reference separated by a space -->
    <xsl:choose>
      <xsl:when test="$posnOfLastSlashCharInSchema != 0">
        <xsl:value-of select="substring($schema, 1, $posnOfLastSlashCharInSchema - 1)"/>
      </xsl:when>
      <xsl:otherwise>
        <xsl:value-of select="$schema"/>
      </xsl:otherwise>
    </xsl:choose>
  </xsl:variable>

  <xsl:variable name="schemaRef" select="concat($schemaPath, ' ', $schema)"/>

  <xsl:call-template name="NewLine"/>
  <xsl:text disable-output-escaping="yes">&lt;gml:FeatureCollection gml:id="id</xsl:text>
    <xsl:for-each select="JOBFile/FieldBook/PointRecord[1]">
      <xsl:call-template name="GenerateGUID"/>
    </xsl:for-each>
    <xsl:text>"</xsl:text>
    <xsl:call-template name="NewLine"/>

    <xsl:text>  xsi:schemaLocation="</xsl:text>
    <xsl:value-of select="$schemaRef"/>
    <xsl:text>"</xsl:text>
    <xsl:call-template name="NewLine"/>

    <xsl:text>  xmlns:gss="http://www.isotc211.org/2005/gss" xmlns:gts="http://www.isotc211.org/2005/gts"</xsl:text>
    <xsl:call-template name="NewLine"/>

    <xsl:text>  xmlns:gsr="http://www.isotc211.org/2005/gsr"</xsl:text>
    <xsl:call-template name="NewLine"/>

    <xsl:text>  xmlns:app="</xsl:text>
    <xsl:value-of select="$schemaPath"/>
    <xsl:text>"</xsl:text>
    <xsl:call-template name="NewLine"/>

    <xsl:text>  xmlns:gco="http://www.isotc211.org/2005/gco" xmlns:gmd="http://www.isotc211.org/2005/gmd"</xsl:text>
    <xsl:call-template name="NewLine"/>

    <xsl:text>  xmlns:xlink="http://www.w3.org/1999/xlink" xmlns:gml="http://www.opengis.net/gml/3.2"</xsl:text>
    <xsl:call-template name="NewLine"/>

    <xsl:text disable-output-escaping="yes">  xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"&gt;</xsl:text>
    <xsl:call-template name="NewLine"/>

    <!-- Output the extents of the job/file -->
    <xsl:call-template name="Indent"/>
    <xsl:text disable-output-escaping="yes">&lt;gml:boundedBy&gt;</xsl:text>
      <xsl:call-template name="NewLine"/>
      <xsl:call-template name="Indent"><xsl:with-param name="level">2</xsl:with-param></xsl:call-template>
      <xsl:text disable-output-escaping="yes">&lt;gml:Envelope srsName="EPSG:5972" srsDimension="3"&gt;</xsl:text>
        <xsl:call-template name="NewLine"/>
        <xsl:call-template name="Indent"><xsl:with-param name="level">3</xsl:with-param></xsl:call-template>
        <xsl:text disable-output-escaping="yes">&lt;gml:lowerCorner&gt;</xsl:text>
          <xsl:value-of select="format-number(msxsl:node-set($coordRange)/minEast, $DecPl3, 'Standard')"/>
          <xsl:value-of select="' '"/>
          <xsl:value-of select="format-number(msxsl:node-set($coordRange)/minNorth, $DecPl3, 'Standard')"/>
          <xsl:value-of select="' '"/>
          <xsl:value-of select="format-number(msxsl:node-set($coordRange)/minElev, $DecPl3, 'Standard')"/>
        <xsl:text disable-output-escaping="yes">&lt;/gml:lowerCorner&gt;</xsl:text>
        <xsl:call-template name="NewLine"/>

        <xsl:call-template name="Indent"><xsl:with-param name="level">3</xsl:with-param></xsl:call-template>
        <xsl:text disable-output-escaping="yes">&lt;gml:upperCorner&gt;</xsl:text>
          <xsl:value-of select="format-number(msxsl:node-set($coordRange)/maxEast, $DecPl3, 'Standard')"/>
          <xsl:value-of select="' '"/>
          <xsl:value-of select="format-number(msxsl:node-set($coordRange)/maxNorth, $DecPl3, 'Standard')"/>
          <xsl:value-of select="' '"/>
          <xsl:value-of select="format-number(msxsl:node-set($coordRange)/maxElev, $DecPl3, 'Standard')"/>
        <xsl:text disable-output-escaping="yes">&lt;/gml:upperCorner&gt;</xsl:text>
        <xsl:call-template name="NewLine"/>
        <xsl:call-template name="Indent"><xsl:with-param name="level">2</xsl:with-param></xsl:call-template>
      <xsl:text disable-output-escaping="yes">&lt;/gml:Envelope&gt;</xsl:text>
      <xsl:call-template name="NewLine"/>
      <xsl:call-template name="Indent"/>
    <xsl:text disable-output-escaping="yes">&lt;/gml:boundedBy&gt;</xsl:text>
    <xsl:call-template name="NewLine"/>

    <!-- Output the data from the Reductions section of the file -->
    <xsl:apply-templates select="JOBFile/Reductions"/>

  <xsl:text disable-output-escaping="yes">&lt;/gml:FeatureCollection&gt;</xsl:text>
  <xsl:call-template name="NewLine"/>

</xsl:template>


<!-- **************************************************************** -->
<!-- ***************** Reductions Node Processing ******************* -->
<!-- **************************************************************** -->
<xsl:template match="JOBFile/Reductions">

  <!-- Output sequenced line points -->
  <xsl:call-template name="OutputLineFeaturePoints"/>
  
  <!-- Output the point elements -->
  <xsl:call-template name="OutputPointFeaturePoints"/>
</xsl:template>


<!-- **************************************************************** -->
<!-- *************** Output the line point sequences **************** -->
<!-- **************************************************************** -->
<xsl:template name="OutputLineFeaturePoints">

  <xsl:variable name="linePoints">
    <xsl:for-each select="Point">
      <xsl:variable name="isKnownLineCode">
        <xsl:call-template name="KnownLineCode"/>
      </xsl:variable>

      <xsl:if test="$isKnownLineCode = 'true'">
        <xsl:element name="Point" namespace="">
          <xsl:copy-of select="./*"/>
          <xsl:element name="lineCode" namespace="">
            <xsl:call-template name="GetLineCode"/>
          </xsl:element>
          <xsl:element name="controlCode" namespace="">
            <xsl:call-template name="GetControlCode"/>
          </xsl:element>
        </xsl:element>
      </xsl:if>
    </xsl:for-each>
  </xsl:variable>

  <xsl:variable name="sortedLinePoints">   <!-- Group like codes together -->
    <xsl:for-each select="msxsl:node-set($linePoints)/Point">
      <xsl:sort data-type="text" order="ascending" select="lineCode"/>  <!-- Sequence by code first -->
      <xsl:sort data-type="text" order="ascending" select="ID"/>        <!-- Then sequence by ID (chronological) -->
      <xsl:copy-of select="."/>
    </xsl:for-each>
  </xsl:variable>

  <xsl:variable name="groupedLinePoints">   <!-- Group into sequences of joined points -->
    <xsl:for-each select="msxsl:node-set($sortedLinePoints)/Point">
      <xsl:variable name="code" select="lineCode"/>

      <!-- Determine if a new group (line joining sequence) needs to be started -->
      <xsl:if test="(position() = 1) or
                    (preceding-sibling::Point[1]/lineCode != $code)  or
                    (preceding-sibling::Point[1]/controlCode = 'End') or
                    (controlCode = 'Start')">
        <xsl:element name="group" namespace="">
          <!-- Copy the current point element -->
          <xsl:copy-of select="."/>
          
          <!-- Now continue on through the Point elements adding to the group until the end of a sequence is encountered -->
          <!-- Grab the following points that have the same code -->
          <xsl:variable name="followingPtsWithSameCode">
            <xsl:for-each select="following-sibling::Point[lineCode = $code]">
              <xsl:copy-of select="."/>
            </xsl:for-each>
          </xsl:variable>

          <!-- Output the following points until the first encountered start or end (outputting the first 'end' point) -->
          <xsl:if test="count(msxsl:node-set($followingPtsWithSameCode)/Point) != 0">
            <xsl:for-each select="msxsl:node-set($followingPtsWithSameCode)/Point">
              <xsl:if test="(count(preceding-sibling::Point[controlCode = 'End']) = 0) and
                            (count(preceding-sibling::Point[controlCode = 'Start']) = 0) and
                            (controlCode != 'Start')">
                <xsl:copy-of select="."/>
              </xsl:if>
            </xsl:for-each>
          </xsl:if>
        </xsl:element>
      </xsl:if>
    </xsl:for-each>
  </xsl:variable>

  <!-- Output grouped line elements -->
  <xsl:for-each select="msxsl:node-set($groupedLinePoints)/group[count(Point) != 0]">
    <!-- Start of a sequence of line points with the same code -->
    <xsl:variable name="code" select="Point[1]/lineCode"/>

    <xsl:variable name="id">
      <xsl:for-each select="Point[1]">
        <xsl:call-template name="GenerateGUID"/>
      </xsl:for-each>
    </xsl:variable>

    <xsl:call-template name="Indent"/>
    <xsl:text disable-output-escaping="yes">&lt;gml:featureMember&gt;</xsl:text>  <!-- Create the gml:FeatureMember element -->
    <xsl:call-template name="NewLine"/>

      <xsl:call-template name="Indent"><xsl:with-param name="level">2</xsl:with-param></xsl:call-template>
      <xsl:text disable-output-escaping="yes">&lt;app:</xsl:text>
      <xsl:value-of select="Point[1]/lineCode"/>
      <xsl:text> gml:id="</xsl:text>
      <xsl:value-of select="concat('_', $id)"/>
      <xsl:text disable-output-escaping="yes">"&gt;</xsl:text>
      <xsl:call-template name="NewLine"/>

        <!-- Output point attributes - based only on the attributes of the first point in the line join sequence -->
        <xsl:variable name="pointAttributes">  <!-- Collect the output attribute lines into a node-set variable -->
          <xsl:call-template name="OutputPointAttributes">
            <xsl:with-param name="point" select="Point[1]"/>
          </xsl:call-template>
        </xsl:variable>

        <xsl:variable name="group" select="."/> <!-- Assign the current group of points to a variable so that it can be passed to the OutputLineSequencePoints template -->
    
        <xsl:for-each select="msxsl:node-set($pointAttributes)/line">
          <xsl:choose>
            <xsl:when test="contains(preceding-sibling::line[1], 'høydereferanse')">
              <!-- The preceding output attribute was the høydereferanse attribue - the line point output needs to follow this -->
              <xsl:call-template name="OutputLineSequencePoints">
                <xsl:with-param name="points" select="$group"/>
                <xsl:with-param name="id" select="$id"/>
              </xsl:call-template>
              <!-- Now output the next line that was going to be output -->
              <xsl:value-of select="." disable-output-escaping="yes"/>
            </xsl:when>
            <xsl:otherwise>
              <xsl:value-of select="." disable-output-escaping="yes"/>
            </xsl:otherwise>
          </xsl:choose>
        </xsl:for-each>
    
      <xsl:call-template name="Indent"><xsl:with-param name="level">2</xsl:with-param></xsl:call-template>
      <xsl:text disable-output-escaping="yes">&lt;/app:</xsl:text>
      <xsl:value-of select="Point[1]/lineCode"/>
      <xsl:text disable-output-escaping="yes">&gt;</xsl:text>
      <xsl:call-template name="NewLine"/>

    <xsl:call-template name="Indent"/>
    <xsl:text disable-output-escaping="yes">&lt;/gml:featureMember&gt;</xsl:text>  <!-- Close the gml:FeatureMember element -->
    <xsl:call-template name="NewLine"/>
  </xsl:for-each>
</xsl:template>


<!-- **************************************************************** -->
<!-- **************** Output the line sequence points *************** -->
<!-- **************************************************************** -->
<xsl:template name="OutputLineSequencePoints">
  <xsl:param name="points"/>
  <xsl:param name="id"/>

  <!-- Output the sequence of point coords for the line sequence -->
  <xsl:call-template name="Indent">
    <xsl:with-param name="level">3</xsl:with-param>
  </xsl:call-template>
  <xsl:text disable-output-escaping="yes">&lt;app:beliggenhet&gt;</xsl:text>
  <xsl:call-template name="NewLine"/>

  <xsl:call-template name="Indent">
    <xsl:with-param name="level">4</xsl:with-param>
  </xsl:call-template>
  <xsl:text disable-output-escaping="yes">&lt;gml:LineString gml:id="</xsl:text>
  <!-- Open element <gml:LineString> -->
  <xsl:value-of select="concat('_', $id, '_geom1')"/>
  <xsl:text disable-output-escaping="yes">" srsName="EPSG:5972" srsDimension="3"&gt;</xsl:text>
  <xsl:call-template name="NewLine"/>

  <xsl:call-template name="Indent">
    <xsl:with-param name="level">5</xsl:with-param>
  </xsl:call-template>
  <xsl:text disable-output-escaping="yes">&lt;gml:posList&gt;</xsl:text>
  <!-- Open element <gml:posList> -->
  <!-- Output the coordinates of the points in the sequence -->
  <xsl:for-each select="msxsl:node-set($points)/Point">
    <xsl:value-of select="format-number(Grid/East, $DecPl3, 'Standard')"/>
    <xsl:value-of select="' '"/>
    <xsl:value-of select="format-number(Grid/North, $DecPl3, 'Standard')"/>
    <xsl:value-of select="' '"/>
    <xsl:value-of select="format-number(Grid/Elevation, $DecPl3, 'Standard')"/>
    <xsl:if test="position() != last()">
      <xsl:value-of select="' '"/>
    </xsl:if>
  </xsl:for-each>
  <xsl:text disable-output-escaping="yes">&lt;/gml:posList&gt;</xsl:text>
  <!-- Close element <gml:posList> -->
  <xsl:call-template name="NewLine"/>

  <xsl:call-template name="Indent">
    <xsl:with-param name="level">4</xsl:with-param>
  </xsl:call-template>
  <xsl:text disable-output-escaping="yes">&lt;/gml:LineString&gt;</xsl:text>
  <!-- Close element <gml:LineString> -->
  <xsl:call-template name="NewLine"/>

  <xsl:call-template name="Indent">
    <xsl:with-param name="level">3</xsl:with-param>
  </xsl:call-template>
  <xsl:text disable-output-escaping="yes">&lt;/app:beliggenhet&gt;</xsl:text>
  <xsl:call-template name="NewLine"/>

</xsl:template>

  
<!-- **************************************************************** -->
<!-- *********** Output the individual point feature points ********* -->
<!-- **************************************************************** -->
<xsl:template name="OutputPointFeaturePoints">

  <xsl:for-each select="Point">
    <xsl:variable name="isKnownPointCode">
      <xsl:call-template name="KnownPointCode"/>
    </xsl:variable>

    <xsl:if test="$isKnownPointCode = 'true'">
      <xsl:variable name="code" select="Code"/>
      <xsl:variable name="north" select="Grid/North"/>
      <xsl:variable name="east" select="Grid/East"/>
      <xsl:variable name="elevation" select="Grid/Elevation"/>
      <xsl:variable name="id">
        <xsl:call-template name="GenerateGUID"/>
      </xsl:variable>

      <xsl:call-template name="Indent"/>
      <xsl:text disable-output-escaping="yes">&lt;gml:featureMember&gt;</xsl:text>  <!-- Open element <gml:featureMember> -->
      <xsl:call-template name="NewLine"/>

        <xsl:call-template name="Indent"><xsl:with-param name="level">2</xsl:with-param></xsl:call-template>
        <xsl:text disable-output-escaping="yes">&lt;app:</xsl:text>
        <xsl:value-of select="Code"/>
        <xsl:text> gml:id="</xsl:text>
        <xsl:value-of select="concat('_', $id)"/>
        <xsl:text disable-output-escaping="yes">"&gt;</xsl:text>
        <xsl:call-template name="NewLine"/>

          <!-- Output point attributes -->
          <xsl:variable name="pointAttributes">  <!-- Collect the output attribute lines into a node-set variable -->
            <xsl:call-template name="OutputPointAttributes">
              <xsl:with-param name="point" select="."/>
              <xsl:with-param name="linePoint">false</xsl:with-param>
            </xsl:call-template>
          </xsl:variable>

          <xsl:for-each select="msxsl:node-set($pointAttributes)/line">
            <xsl:choose>
              <xsl:when test="contains(following-sibling::line[1], 'identifikasjon') and (count(preceding-sibling::line[contains(., 'identifikasjon')]) = 0)">
                <xsl:value-of select="." disable-output-escaping="yes"/>  <!-- Output the current attribute line -->

                <!-- The next output line is the starting line of the identifikasjon attribue - the point position output needs to precede this -->
                <!-- Output the point coordinates -->
                <xsl:call-template name="OutputPointCoordinates">
                  <xsl:with-param name="id" select="$id"/>
                  <xsl:with-param name="code" select="$code"/>
                  <xsl:with-param name="north" select="$north"/>
                  <xsl:with-param name="east" select="$east"/>
                  <xsl:with-param name="elevation" select="$elevation"/>
                </xsl:call-template>
              </xsl:when>
              <xsl:otherwise>
                <xsl:value-of select="." disable-output-escaping="yes"/>
              </xsl:otherwise>
            </xsl:choose>
          </xsl:for-each>

        <xsl:call-template name="Indent"><xsl:with-param name="level">2</xsl:with-param></xsl:call-template>
        <xsl:text disable-output-escaping="yes">&lt;/app:</xsl:text>
        <xsl:value-of select="Code"/>
        <xsl:text disable-output-escaping="yes">&gt;</xsl:text>
        <xsl:call-template name="NewLine"/>

      <xsl:call-template name="Indent"/>
      <xsl:text disable-output-escaping="yes">&lt;/gml:featureMember&gt;</xsl:text>  <!-- Close element <gml:featureMember> -->
      <xsl:call-template name="NewLine"/>
    </xsl:if>
  </xsl:for-each>
</xsl:template>


<!-- **************************************************************** -->
<!-- ******** Output the attributes for the passed in point ********* -->
<!-- **************************************************************** -->
<xsl:template name="OutputPointAttributes">
  <xsl:param name="point"/>
  <xsl:param name="linePoint" select="'true'"/>

  <!-- Outputs the point attributes that have values assigned to them in the order that they are recorded in the JobXML file -->
  <xsl:for-each select="msxsl:node-set($point)/Features/Feature">
    <xsl:variable name="hasAttributes">
      <xsl:call-template name="AttributeDefinitionAvailable">
        <xsl:with-param name="code" select="@Name"/>
        <xsl:with-param name="linePoint" select="$linePoint"/>
      </xsl:call-template>
    </xsl:variable>
    
    <xsl:if test="$hasAttributes = 'true'">  <!-- Only continue if there are attributes to output -->
      <xsl:variable name="featureName" select="@Name"/>
      <xsl:for-each select="Attribute">
        <xsl:if test="(Value != '') and (Value != '?')">  <!-- Attribute has a value -->
          <!-- Get the attribute definition for this attribute -->
          <xsl:variable name="thisAttribDefinition">
            <xsl:call-template name="GetAttributeDefinition">
              <xsl:with-param name="featureName" select="$featureName"/>
              <xsl:with-param name="attribName" select="Name"/>
              <xsl:with-param name="linePoint" select="$linePoint"/>
            </xsl:call-template>
          </xsl:variable>

          <xsl:variable name="thisAttribPath">  <!-- Path up to but not including the element itself -->
            <xsl:variable name="pathCount" select="count(msxsl:node-set($thisAttribDefinition)/attribute/inPath/path)"/>
            <xsl:for-each select="msxsl:node-set($thisAttribDefinition)/attribute/inPath/path">
              <xsl:if test="position() &lt; $pathCount">
                <xsl:value-of select="."/>
                <xsl:text>/</xsl:text>
              </xsl:if>
            </xsl:for-each>
          </xsl:variable>
                        
          <!-- Get the attribute definition for the preceding non-empty attribute -->
          <xsl:variable name="prevAttribDefinition">
            <xsl:call-template name="GetAttributeDefinition">
              <xsl:with-param name="featureName" select="$featureName"/>
              <xsl:with-param name="attribName" select="preceding-sibling::Attribute[(Value != '') and (Value != '?')][1]/Name"/>
              <xsl:with-param name="linePoint" select="$linePoint"/>
            </xsl:call-template>
          </xsl:variable>
          
          <xsl:variable name="prevAttribPath">  <!-- Path up to but not including the element itself -->
            <xsl:variable name="pathCount" select="count(msxsl:node-set($prevAttribDefinition)/attribute/inPath/path)"/>
            <xsl:for-each select="msxsl:node-set($prevAttribDefinition)/attribute/inPath/path">
              <xsl:if test="position() &lt; $pathCount">
                <xsl:value-of select="."/>
                <xsl:text>/</xsl:text>
              </xsl:if>
            </xsl:for-each>
          </xsl:variable>

          <xsl:variable name="attribValue" select="Value"/>  <!-- Hold the attribute value so that it is available in the for-each path loop -->
        
          <xsl:choose>
            <xsl:when test="string($prevAttribPath) = ''">
              <!-- No preceding attribute path - need to build the current path and output the attribute value -->
              <xsl:for-each select="msxsl:node-set($thisAttribDefinition)/attribute/inPath/path">
                <xsl:element name="line">
                  <xsl:call-template name="Indent"><xsl:with-param name="level" select="2 + position()"/></xsl:call-template>
                  <xsl:value-of select="concat('&lt;', ., '&gt;')" disable-output-escaping="yes"/>
                  <xsl:if test="position() = last()">
                    <!-- Now output the attribute value -->
                    <xsl:value-of select="$attribValue"/>
                    <!-- And 'close' the element -->
                    <xsl:value-of select="concat('&lt;/', ., '&gt;')" disable-output-escaping="yes"/>
                  </xsl:if>
                  <xsl:call-template name="NewLine"/>
                </xsl:element>  
              </xsl:for-each>
            </xsl:when>

            <xsl:when test="string($thisAttribPath) = string($prevAttribPath)">  <!-- Same path as preceding non-empty attribute -->
              <!-- Only need to output the actual value element itself -->
              <xsl:element name="line">
                <xsl:call-template name="Indent"><xsl:with-param name="level" select="count(msxsl:node-set($thisAttribDefinition)/attribute/inPath/path) + 2"/></xsl:call-template>
                <xsl:value-of select="concat('&lt;', msxsl:node-set($thisAttribDefinition)/attribute/inPath/path[last()], '&gt;')" disable-output-escaping="yes"/>
                <!-- Now output the attribute value -->
                <xsl:value-of select="$attribValue"/>
                <!-- And 'close' the element -->
                <xsl:value-of select="concat('&lt;/', msxsl:node-set($thisAttribDefinition)/attribute/inPath/path[last()], '&gt;')" disable-output-escaping="yes"/>
                <xsl:call-template name="NewLine"/>
              </xsl:element>
            </xsl:when>
            
            <xsl:when test="string($thisAttribPath) != string($prevAttribPath)">  <!-- Different path from the preceding non-empty attribute -->
              <!-- First need to close off the previous element path -->
              <xsl:variable name="pathCount" select="count(msxsl:node-set($prevAttribDefinition)/attribute/outPath/path) - 1"/>
              <xsl:for-each select="msxsl:node-set($prevAttribDefinition)/attribute/outPath/path[position() &gt; 1]">
                <xsl:element name="line">
                  <xsl:call-template name="Indent"><xsl:with-param name="level" select="3 + $pathCount - position()"/></xsl:call-template>
                  <xsl:value-of select="concat('&lt;/', ., '&gt;')" disable-output-escaping="yes"/>
                  <xsl:call-template name="NewLine"/>
                </xsl:element>
              </xsl:for-each>
              
              <!-- Now build the current path and output the attribute value -->
              <xsl:for-each select="msxsl:node-set($thisAttribDefinition)/attribute/inPath/path">
                <xsl:element name="line">
                  <xsl:call-template name="Indent"><xsl:with-param name="level" select="2 + position()"/></xsl:call-template>
                  <xsl:value-of select="concat('&lt;', ., '&gt;')" disable-output-escaping="yes"/>
                  <xsl:if test="position() = last()">
                    <!-- Now output the attribute value -->
                    <xsl:value-of select="$attribValue"/>
                    <!-- And 'close' the element -->
                    <xsl:value-of select="concat('&lt;/', ., '&gt;')" disable-output-escaping="yes"/>
                  </xsl:if>
                  <xsl:call-template name="NewLine"/>
                </xsl:element>
              </xsl:for-each>
            </xsl:when>
          </xsl:choose>

          <xsl:if test="count(following-sibling::Attribute[(Value != '') and (Value != '?')]) = 0">
            <!-- There are no further attributes with values assigned to them - 'close off' the current element if required -->
            <xsl:if test="count(msxsl:node-set($thisAttribDefinition)/attribute/outPath/path) &gt; 1">
              <xsl:variable name="pathCount" select="count(msxsl:node-set($thisAttribDefinition)/attribute/outPath/path) - 1"/>
              <xsl:for-each select="msxsl:node-set($thisAttribDefinition)/attribute/outPath/path[position() &gt; 1]">
                <xsl:element name="line">
                  <xsl:call-template name="Indent"><xsl:with-param name="level" select="3 + $pathCount - position()"/></xsl:call-template>
                  <xsl:value-of select="concat('&lt;/', ., '&gt;')" disable-output-escaping="yes"/>
                  <xsl:call-template name="NewLine"/>
                </xsl:element>
              </xsl:for-each>
            </xsl:if>
          </xsl:if>
        </xsl:if>
      </xsl:for-each>
    </xsl:if>
  </xsl:for-each>
</xsl:template>


<!-- **************************************************************** -->
<!-- ***** Check that attrib definitions are available for code ***** -->
<!-- **************************************************************** -->
<xsl:template name="AttributeDefinitionAvailable">
  <xsl:param name="code"/>
  <xsl:param name="linePoint" select="'true'"/>

  <xsl:choose>
    <xsl:when test="$linePoint = 'true'">
      <xsl:choose>
        <xsl:when test="count(msxsl:node-set($lineCodes)/code[name = $code]) != 0">true</xsl:when>
        <xsl:otherwise>false</xsl:otherwise>
      </xsl:choose>
    </xsl:when>
    <xsl:otherwise>
      <xsl:choose>
        <xsl:when test="count(msxsl:node-set($pointCodes)/code[name = $code]) != 0">true</xsl:when>
        <xsl:otherwise>false</xsl:otherwise>
      </xsl:choose>
    </xsl:otherwise>
  </xsl:choose>
</xsl:template>


<!-- **************************************************************** -->
<!-- ******* Get the attrib path from the attrib definitions ******** -->
<!-- **************************************************************** -->
<xsl:template name="GetAttributeDefinition">
  <xsl:param name="featureName"/>
  <xsl:param name="attribName"/>
  <xsl:param name="linePoint" select="'true'"/>

  <xsl:choose>
    <xsl:when test="$linePoint = 'true'">
      <xsl:for-each select="msxsl:node-set($lineCodes)/code[name = $featureName]/attributes/attribute[@name = $attribName]">
        <xsl:copy-of select="."/>
      </xsl:for-each>
    </xsl:when>
    <xsl:otherwise>
      <xsl:for-each select="msxsl:node-set($pointCodes)/code[name = $featureName]/attributes/attribute[@name = $attribName]">
        <xsl:copy-of select="."/>
      </xsl:for-each>
    </xsl:otherwise>
  </xsl:choose>
</xsl:template>


<!-- **************************************************************** -->
<!-- ********* Extract the path values from combined string ********* -->
<!-- **************************************************************** -->
<xsl:template name="ExtractPathElements">
  <xsl:param name="pathString"/>

  <xsl:if test="string-length($pathString) != 0">
    <xsl:choose>
      <!-- No prefixes set for this attribute so set path to the attribute prefixed with app: -->  
      <xsl:when test="count(msxsl:node-set($attribPrefixes)/*[$pathString = name()]) = 0">
        <xsl:element name="path" namespace="">
          <xsl:text>app:</xsl:text>
          <xsl:value-of select="$pathString"/>
        </xsl:element>
      </xsl:when>
      <xsl:otherwise>
        <xsl:for-each select="msxsl:node-set($attribPrefixes)/*">
          <xsl:if test="$pathString = name()">
            <!-- Add in a path element for each attribute prefix -->  
            <xsl:for-each select="./prefix">
              <xsl:element name="path" namespace="">
                <xsl:text>app:</xsl:text>
                <xsl:value-of select="."/>
              </xsl:element>
            </xsl:for-each>
            <!-- Now add the actual attribute name as the last 'path' -->
            <xsl:element name="path" namespace="">
              <xsl:text>app:</xsl:text>
              <xsl:value-of select="$pathString"/>
            </xsl:element>
          </xsl:if>
        </xsl:for-each>
      </xsl:otherwise>  
    </xsl:choose>  
  </xsl:if>
</xsl:template>


<!-- **************************************************************** -->
<!-- ******** Test to see if current point code is in library ******* -->
<!-- **************************************************************** -->
<xsl:template name="KnownPointCode">

  <xsl:variable name="code" select="Code"/>  <!-- Hold the code for the current point in variable -->
  <xsl:choose>
    <xsl:when test="count(msxsl:node-set($pointCodes)/code/name[. = $code]) != 0">true</xsl:when>
    <xsl:otherwise>false</xsl:otherwise>
  </xsl:choose>
</xsl:template>


<!-- **************************************************************** -->
<!-- ******** Test to see if current line code is in library ******** -->
<!-- **************************************************************** -->
<xsl:template name="KnownLineCode">

  <xsl:variable name="code" select="Code"/>  <!-- Hold the code for the current point in variable -->
  <xsl:choose>
    <xsl:when test="count(msxsl:node-set($lineCodes)/code/name[contains($code, .)]) != 0">true</xsl:when>  <!-- Allow for line codes with 'counter' -->
    <xsl:otherwise>false</xsl:otherwise>
  </xsl:choose>
</xsl:template>


<!-- **************************************************************** -->
<!-- ******* Return the first line code found in Code element ******* -->
<!-- **************************************************************** -->
<xsl:template name="GetLineCode">

  <xsl:variable name="code" select="Code"/>
  <xsl:variable name="codes">
    <xsl:for-each select="msxsl:node-set($lineCodes)/code/name[contains($code, .)]">
      <xsl:element name="code" namespace="">
        <xsl:value-of select="."/>
      </xsl:element>
    </xsl:for-each>
  </xsl:variable>
  
  <xsl:value-of select="msxsl:node-set($codes)/code[1]"/>
</xsl:template>


<!-- **************************************************************** -->
<!-- ****** Return the first control code found in Code element ****** -->
<!-- **************************************************************** -->
<xsl:template name="GetControlCode">

  <xsl:variable name="code" select="Code"/>
  <xsl:variable name="codes">
    <xsl:for-each select="msxsl:node-set($controlCodes)/code[contains($code, .)]">
      <xsl:element name="code" namespace="">
        <xsl:value-of select="./@type"/>
      </xsl:element>
    </xsl:for-each>
  </xsl:variable>

  <xsl:value-of select="msxsl:node-set($codes)/code[1]"/>
</xsl:template>


<!-- **************************************************************** -->
<!-- ******************* Output Point Coordinates ******************* -->
<!-- **************************************************************** -->
<xsl:template name="OutputPointCoordinates">
  <xsl:param name="baseIndent" select="3"/>
  <xsl:param name="id"/>
  <xsl:param name="code"/>
  <xsl:param name="north"/>
  <xsl:param name="east"/>
  <xsl:param name="elevation"/>

  <xsl:call-template name="Indent"><xsl:with-param name="level" select="$baseIndent"/></xsl:call-template>
  <xsl:choose>
    <xsl:when test="$code = 'Kumlokk'">
      <xsl:text disable-output-escaping="yes">&lt;app:senterLokk&gt;</xsl:text>
    </xsl:when>
    <xsl:otherwise>
      <xsl:text disable-output-escaping="yes">&lt;app:posisjon&gt;</xsl:text>
    </xsl:otherwise>
  </xsl:choose>
    <xsl:call-template name="NewLine"/>
    <xsl:call-template name="Indent"><xsl:with-param name="level" select="$baseIndent + 1"/></xsl:call-template>
    <xsl:text disable-output-escaping="yes">&lt;gml:Point gml:id="</xsl:text>  <!-- Open element <gml:Point> -->
    <xsl:value-of select="concat('_', $id, '_geom1')"/>
    <xsl:text disable-output-escaping="yes">" srsName="EPSG:5972" srsDimension="3"&gt;</xsl:text>
    <xsl:call-template name="NewLine"/>

      <xsl:call-template name="Indent"><xsl:with-param name="level" select="$baseIndent + 2"/></xsl:call-template>
      <xsl:text disable-output-escaping="yes">&lt;gml:pos&gt;</xsl:text>  <!-- Open element <gml:pos> -->
        <xsl:value-of select="format-number($east, $DecPl3, 'Standard')"/>
        <xsl:value-of select="' '"/>
        <xsl:value-of select="format-number($north, $DecPl3, 'Standard')"/>
        <xsl:value-of select="' '"/>
        <xsl:value-of select="format-number($elevation, $DecPl3, 'Standard')"/>
      <xsl:text disable-output-escaping="yes">&lt;/gml:pos&gt;</xsl:text>  <!-- Close element <gml:pos> -->
      <xsl:call-template name="NewLine"/>
    <xsl:call-template name="Indent"><xsl:with-param name="level" select="$baseIndent + 1"/></xsl:call-template>
    <xsl:text disable-output-escaping="yes">&lt;/gml:Point&gt;</xsl:text>  <!-- Close element <gml:Point> -->
    <xsl:call-template name="NewLine"/>
    <xsl:call-template name="Indent"><xsl:with-param name="level" select="$baseIndent"/></xsl:call-template>
  <xsl:choose>
    <xsl:when test="$code = 'Kumlokk'">
      <xsl:text disable-output-escaping="yes">&lt;/app:senterLokk&gt;</xsl:text>
    </xsl:when>
    <xsl:otherwise>
      <xsl:text disable-output-escaping="yes">&lt;/app:posisjon&gt;</xsl:text>
    </xsl:otherwise>
  </xsl:choose>

  <xsl:call-template name="NewLine"/>
</xsl:template>


<!-- **************************************************************** -->
<!-- ********************* Generate random GUID ********************* -->
<!-- **************************************************************** -->
<xsl:template name="GenerateGUID">

  <!-- A routine to generate a 'reasonably random' GUID type value using the available generate-id function. -->
  <xsl:variable name="seg1" select="generate-id(.)"/>
  <xsl:variable name="seg2" select="generate-id(./Name)"/>
  <xsl:variable name="seg3" select="generate-id(./Code)"/>
  <xsl:variable name="seg4" select="generate-id(./Method)"/>
  <xsl:variable name="seg5" select="generate-id(./SurveyMethod)"/>
  <xsl:variable name="seg6" select="generate-id(./Classification)"/>

  <xsl:variable name="initStr" select="translate(substring(concat($seg1, $seg2, $seg3, $seg4, $seg5, $seg6), 1, 32),'ABCDEFGHIJKLMNOPQRSTUVWXYZ','abcdefghijklmnopqrstuvwxyz')"/>
  
  <xsl:variable name="hexStr">
    <xsl:call-template name="ConvertToHex">
      <xsl:with-param name="inStr" select="$initStr"/>
    </xsl:call-template>
  </xsl:variable>
  
  <xsl:value-of select="substring($hexStr, 1, 8)"/>
  <xsl:text>-</xsl:text>
  <xsl:value-of select="substring($hexStr, 9, 4)"/>
  <xsl:text>-</xsl:text>
  <xsl:value-of select="substring($hexStr, 13, 4)"/>
  <xsl:text>-</xsl:text>
  <xsl:value-of select="substring($hexStr, 17, 4)"/>
  <xsl:text>-</xsl:text>
  <xsl:value-of select="substring($hexStr, 21, 12)"/>
</xsl:template>


<!-- **************************************************************** -->
<!-- *************** Convert to string of hex values **************** -->
<!-- **************************************************************** -->
<xsl:template name="ConvertToHex">
  <xsl:param name="inStr"/>
  <xsl:param name="posn" select="1"/>

  <xsl:choose>
    <xsl:when test="$posn &lt;= string-length($inStr)">
      <xsl:variable name="char" select="substring($inStr, $posn, 1)"/>
      <xsl:variable name="newChar">
        <xsl:choose>
          <xsl:when test="$char = '0'">0</xsl:when>
          <xsl:when test="$char = '1'">1</xsl:when>
          <xsl:when test="$char = '2'">2</xsl:when>
          <xsl:when test="$char = '3'">3</xsl:when>
          <xsl:when test="$char = '4'">4</xsl:when>
          <xsl:when test="$char = '5'">5</xsl:when>
          <xsl:when test="$char = '6'">6</xsl:when>
          <xsl:when test="$char = '7'">7</xsl:when>
          <xsl:when test="$char = '8'">8</xsl:when>
          <xsl:when test="$char = '9'">9</xsl:when>
          <xsl:when test="$char = 'a'">a</xsl:when>
          <xsl:when test="$char = 'b'">b</xsl:when>
          <xsl:when test="$char = 'c'">c</xsl:when>
          <xsl:when test="$char = 'd'">d</xsl:when>
          <xsl:when test="$char = 'e'">e</xsl:when>
          <xsl:when test="$char = 'f'">f</xsl:when>
          <xsl:when test="$char = 'g'">0</xsl:when>
          <xsl:when test="$char = 'h'">1</xsl:when>
          <xsl:when test="$char = 'i'">2</xsl:when>
          <xsl:when test="$char = 'j'">3</xsl:when>
          <xsl:when test="$char = 'k'">4</xsl:when>
          <xsl:when test="$char = 'l'">5</xsl:when>
          <xsl:when test="$char = 'm'">6</xsl:when>
          <xsl:when test="$char = 'n'">7</xsl:when>
          <xsl:when test="$char = 'o'">8</xsl:when>
          <xsl:when test="$char = 'p'">9</xsl:when>
          <xsl:when test="$char = 'q'">a</xsl:when>
          <xsl:when test="$char = 'r'">b</xsl:when>
          <xsl:when test="$char = 's'">c</xsl:when>
          <xsl:when test="$char = 't'">d</xsl:when>
          <xsl:when test="$char = 'u'">e</xsl:when>
          <xsl:when test="$char = 'v'">f</xsl:when>
          <xsl:when test="$char = 'w'">0</xsl:when>
          <xsl:when test="$char = 'x'">1</xsl:when>
          <xsl:when test="$char = 'y'">2</xsl:when>
          <xsl:when test="$char = 'z'">3</xsl:when>
        </xsl:choose>
      </xsl:variable>

      <xsl:call-template name="ConvertToHex">
        <xsl:with-param name="inStr" select="concat(substring($inStr, 1, $posn - 1), $newChar, substring($inStr, $posn + 1))"/>
        <xsl:with-param name="posn" select="$posn + 1"/>
      </xsl:call-template>
    </xsl:when>

    <xsl:otherwise>
      <xsl:value-of select="$inStr"/>
    </xsl:otherwise>
  </xsl:choose>
</xsl:template>


<!-- **************************************************************** -->
<!-- **************** Indent lines appropriately ******************** -->
<!-- **************************************************************** -->
<xsl:template name="Indent">
  <xsl:param name="level" select="1"/>

  <xsl:variable name="nbrSpaces" select="$level * 2"/>

  <xsl:value-of select="substring('                                       ', 1, $nbrSpaces)"/>
</xsl:template>


<!-- **************************************************************** -->
<!-- **** Return position of the last of specified char in string *** -->
<!-- **************************************************************** -->
<xsl:template name="PosnOfLastOfSpecifiedChars">
  <xsl:param name="stringToSearch"/>
  <xsl:param name="charToLocate"/>
  
  <xsl:choose>
    <xsl:when test="string-length($stringToSearch) = 0">0</xsl:when>  <!-- Nothing to search - return 0 -->

    <xsl:when test="substring($stringToSearch, string-length($stringToSearch), 1) = $charToLocate">
      <xsl:value-of select="string-length($stringToSearch)"/>
    </xsl:when>

    <xsl:otherwise>
      <xsl:choose>
        <xsl:when test="string-length($stringToSearch) = 1">0</xsl:when>  <!-- Down to the last character in string and it is not the one we are looking for - return 0 -->
        <xsl:otherwise>
          <!-- Recurse the function stripping off the last character in the search string -->
          <xsl:call-template name="PosnOfLastOfSpecifiedChars">
            <xsl:with-param name="stringToSearch" select="substring($stringToSearch, 1, string-length($stringToSearch) - 1)"/>
            <xsl:with-param name="charToLocate" select="$charToLocate"/>
          </xsl:call-template>
        </xsl:otherwise>
      </xsl:choose>
    </xsl:otherwise>
  </xsl:choose>
</xsl:template>


<!-- **************************************************************** -->
<!-- ********************** New Line Output ************************* -->
<!-- **************************************************************** -->
<xsl:template name="NewLine">
  <xsl:text>&#10;</xsl:text>
</xsl:template>


<!-- **************************************************************** -->
<!-- ********** Return a Reverse Order Node Set Variable ************ -->
<!-- **************************************************************** -->
<xsl:template name="ReversedNodeSet">
  <xsl:param name="originalNodeSet"/>
  <xsl:param name="count"/>   <!-- Pass in the count of the elements in the node set -->
  <xsl:param name="item"/>    <!-- Initially set this equal to the element count     -->

  <!-- This recursive function will return the passed in node set in the reverse order -->
  <xsl:if test="$item &gt; 0">
    <xsl:choose>
      <xsl:when test="$item = $count">
        <xsl:for-each select="msxsl:node-set($originalNodeSet)/*[last()]">  <!-- Get the last element (returned first) -->
          <xsl:copy>
            <xsl:copy-of select="* | @*"/>
            <xsl:if test="text()">
              <xsl:value-of select="."/>
             </xsl:if>
          </xsl:copy>
        </xsl:for-each>
      </xsl:when>

      <xsl:otherwise>  <!-- Copy the appropriate preceding element -->
        <xsl:for-each select="msxsl:node-set($originalNodeSet)/*[last()]">  <!-- Get the last element -->
          <xsl:for-each select="preceding-sibling::*[$count - $item]">      <!-- get the required preceding element -->
            <xsl:copy>
              <xsl:copy-of select="* | @*"/>
              <xsl:if test="text()">
                <xsl:value-of select="."/>
               </xsl:if>
            </xsl:copy>
          </xsl:for-each>
        </xsl:for-each>
      </xsl:otherwise>
    </xsl:choose>

    <!-- Recurse the function decrementing the item value -->
    <xsl:call-template name="ReversedNodeSet">
      <xsl:with-param name="originalNodeSet" select="$originalNodeSet"/>
      <xsl:with-param name="count" select="$count"/>
      <xsl:with-param name="item" select="$item - 1"/>
    </xsl:call-template>
  </xsl:if>

</xsl:template>


</xsl:stylesheet>